/*
* Copyright (c) 2008-2009 Christian Buchner <Christian.Buchner@gmail.com>
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY Christian Buchner ''AS IS'' AND ANY 
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL Christian Buchner BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <vector_types.h>
#include <stdio.h>

#define STACKHEIGHT 64

#ifdef __CUDACC__
#define HOST __host__
#define DEVICE __device__
#define HOSTDEVICE __host__ __device__
#define M_HOST __host__
#define M_HOSTDEVICE __host__ __device__
#else
#define HOST static inline
#define DEVICE static inline
#define HOSTDEVICE static inline
#define M_HOST inline
#define M_HOSTDEVICE inline
#endif

#ifdef __CUDACC__
#define ALIGN(x)  __align__(x)
#else
#if defined(_MSC_VER) && (_MSC_VER >= 1300)
#define ALIGN(x) __declspec(align(x)) 
#else
#if defined(__GNUC__)
#define ALIGN(x)  __attribute__ ((aligned (x)))
#else
#define ALIGN(x) 
#endif
#endif
#endif

#if !defined(__DEVICE_EMULATION__) || (defined(_MSC_VER) && (_MSC_VER >= 1300))
#define REF(x) &x
#define ARRAYREF(x,y) (&x)[y]
#else
#define REF(x) x
#define ARRAYREF(x,y) x[y]
#endif

typedef struct ALIGN(8) _cudacomplex {

  float2 value;

  _cudacomplex& operator=(const float REF(a)) {
     value.x = a; value.y = 0;
     return *this;
  };

  _cudacomplex& operator=(const float ARRAYREF(a,2)) {
     value.x = a[0]; value.y = a[1];
     return *this;
  };

  M_HOSTDEVICE float& real() {return value.x;};
  M_HOSTDEVICE float& imag() {return value.y;};

} cudacomplex;

HOSTDEVICE cudacomplex operator+(const cudacomplex REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a.value.x + b.value.x, a.value.y  + b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator+(const cudacomplex REF(a), const float REF(b)) {
   cudacomplex result = {{ a.value.x + b, a.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator+(const float REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a + b.value.x, b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator-(const cudacomplex REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a.value.x - b.value.x, a.value.y  - b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator-(const cudacomplex REF(a)) {
   cudacomplex result = {{ -a.value.x, -a.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator-(const cudacomplex REF(a), const float REF(b)) {
   cudacomplex result = {{ a.value.x - b, a.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator-(const float REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a - b.value.x, -b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator*(const cudacomplex REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a.value.x * b.value.x - a.value.y * b.value.y,
                           a.value.y * b.value.x + a.value.x * b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator*(const cudacomplex REF(a), const float REF(b)) {
   cudacomplex result = {{ a.value.x * b, a.value.y * b }};
   return result;
}

HOSTDEVICE cudacomplex operator*(const float REF(a), const cudacomplex REF(b)) {
   cudacomplex result = {{ a * b.value.x, a * b.value.y }};
   return result;
}

HOSTDEVICE cudacomplex operator/(const cudacomplex REF(a), const cudacomplex REF(b)) {
   float tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
   cudacomplex result = {{ (a.value.x * b.value.x + a.value.y * b.value.y ) / tmp,
                           (a.value.y * b.value.x - a.value.x * b.value.y ) / tmp }};
   return result;
}

HOSTDEVICE cudacomplex operator/(const cudacomplex REF(a), const float REF(b)) {
   cudacomplex result = {{ a.value.x / b, a.value.y / b }};
   return result;
}

HOSTDEVICE cudacomplex operator/(const float REF(a), const cudacomplex REF(b)) {
   float tmp = ( b.value.x * b.value.x + b.value.y * b.value.y );
   cudacomplex result = {{ ( a * b.value.x ) / tmp, ( -a * b.value.y ) / tmp }};
   return result;
}

HOSTDEVICE cudacomplex operator~(const cudacomplex REF(a)) {
   cudacomplex result = {{ a.value.x, -a.value.y }};
   return result;
}

HOSTDEVICE cudacomplex make_cudacomplex(float a, float b)
{
    cudacomplex res;
    res.real() = a;
    res.imag() = b;
    return res;
}

namespace constants
{
    const cudacomplex zero = make_cudacomplex(0.0f, 0.0f);
    const cudacomplex one  = make_cudacomplex(1.0f, 0.0f);
    const cudacomplex I    = make_cudacomplex(0.0f, 1.0f);
};

struct matrix2_2 {
    cudacomplex a, b, c, d;

    M_HOSTDEVICE matrix2_2() {
        a = b = c = d = 0.0f;
    }
    M_HOSTDEVICE matrix2_2(const cudacomplex REF(_a), const cudacomplex REF(_b), const cudacomplex REF(_c), const cudacomplex REF(_d)) :
        a(_a), b(_b), c(_c), d(_d) {}

    M_HOSTDEVICE matrix2_2 operator*(const matrix2_2 REF(A)) const {
        return matrix2_2(
            A.a * a + A.c * b, A.b * a + A.d * b,
            A.a * c + A.c * d, A.b * c + A.d * d );
    }

    M_HOSTDEVICE matrix2_2 operator/(const matrix2_2 REF(A)) const {
        matrix2_2 A_inv = A.inv();
        return matrix2_2(
            A_inv.a * a + A_inv.c * b,  A_inv.b * a + A_inv.d * b,
            A_inv.a * c + A_inv.c * d,  A_inv.b * c + A_inv.d * d );
    }

    M_HOSTDEVICE matrix2_2 operator+(const matrix2_2 REF(A)) const {
        return matrix2_2(A.a + a,  A.b + b,
                         A.c + c,  A.d + d);
    }

    M_HOSTDEVICE matrix2_2 operator-(const matrix2_2 REF(A)) const {
        return matrix2_2(a - A.a,  b - A.b,
                         c - A.c,  d - A.d);
    }

    M_HOSTDEVICE matrix2_2 operator-() const {
        return matrix2_2(-a, -b,
                         -c, -d);
    }

    M_HOSTDEVICE cudacomplex det() const {
        return a*d-b*c;
    }

    M_HOSTDEVICE matrix2_2 inv() const {
        cudacomplex  i_det = 1.0f / det();
        cudacomplex _i_det = -i_det;
        return matrix2_2( d *  i_det,  b * _i_det,
                          c * _i_det,  a *  i_det);
    }

    M_HOST void print() {
        printf("[ %.4f + %.4fj\t", a.real(), a.imag()); printf("%.4f + %.4fj\n",   b.real(), b.imag());
        printf("  %.4f + %.4fj\t", c.real(), c.imag()); printf("%.4f + %.4fj ]\n", d.real(), d.imag());
    }

    static M_HOSTDEVICE matrix2_2 zero(void) {
        return matrix2_2(constants::zero, constants::zero,
                         constants::zero, constants::zero);
    }

    static M_HOSTDEVICE matrix2_2 ones(void) {
        return matrix2_2(constants::one,  constants::one,
                         constants::one,  constants::one);
    }

    static M_HOSTDEVICE matrix2_2 unit(void) {
        return matrix2_2(constants::one,  constants::zero,
                         constants::zero, constants::one);
    }
};


template <int height> struct matrixstack2_2
{
    cudacomplex a[height], b[height], c[height], d[height];

    M_HOSTDEVICE matrix2_2 operator[](int index) const {
        return matrix2_2(a[index],b[index],
                         c[index],d[index]);
    }

    M_HOSTDEVICE void set(int index, const matrix2_2 REF(data))
    {
        a[index] = data.a; b[index] = data.b;
        c[index] = data.c; d[index] = data.d;
    }

    M_HOSTDEVICE void set(int index, const cudacomplex REF(_a), const cudacomplex REF(_b), const cudacomplex REF(_c), const cudacomplex REF(_d))
    {
        a[index] = _a; b[index] = _b;
        c[index] = _c; d[index] = _d;
    }
};

struct matrix4_4 {

    matrix2_2 A, B, C, D;

    M_HOSTDEVICE matrix4_4() {
        A = B = C = D = matrix2_2::zero();
    }
    M_HOSTDEVICE matrix4_4(const matrix2_2 REF(_A), const matrix2_2 REF(_B), const matrix2_2 REF(_C), const matrix2_2 REF(_D)) :
        A(_A), B(_B), C(_C), D(_D) {}

    M_HOSTDEVICE matrix4_4 operator*(const matrix4_4 REF(M)) const {

        matrix2_2 _A(A.a*M.A.a + A.b*M.A.c + B.a*M.C.a + B.b*M.C.c,
                     A.a*M.A.b + A.b*M.A.d + B.a*M.C.b + B.b*M.C.d,
                     A.c*M.A.a + A.d*M.A.c + B.c*M.C.a + B.d*M.C.c,
                     A.c*M.A.b + A.d*M.A.d + B.c*M.C.b + B.d*M.C.d);
                    
        matrix2_2 _B(A.a*M.B.a + A.b*M.B.c + B.a*M.D.a + B.b*M.D.c,
                     A.a*M.B.b + A.b*M.B.d + B.a*M.D.b + B.b*M.D.d,
                     A.c*M.B.a + A.d*M.B.c + B.c*M.D.a + B.d*M.D.c,
                     A.c*M.B.b + A.d*M.B.d + B.c*M.D.b + B.d*M.D.d);

        matrix2_2 _C(C.a*M.A.a + C.b*M.A.c + D.a*M.C.a + D.b*M.C.c,
                     C.a*M.A.b + C.b*M.A.d + D.a*M.C.b + D.b*M.C.d,
                     C.c*M.A.a + C.d*M.A.c + D.c*M.C.a + D.d*M.C.c,
                     C.c*M.A.b + C.d*M.A.d + D.c*M.C.b + D.d*M.C.d);

        matrix2_2 _D(C.a*M.B.a + C.b*M.B.c + D.a*M.D.a + D.b*M.D.c,
                     C.a*M.B.b + C.b*M.B.d + D.a*M.D.b + D.b*M.D.d,
                     C.c*M.B.a + C.d*M.B.c + D.c*M.D.a + D.d*M.D.c,
                     C.c*M.B.b + C.d*M.B.d + D.c*M.D.b + D.d*M.D.d);
        
        return matrix4_4(_A,_B,_C,_D);
    }

    M_HOSTDEVICE matrix4_4 inv() const {
        matrix2_2 Ainv = A.inv();
        matrix2_2 T = (D-(C*Ainv*B)).inv();
        return matrix4_4(
                           Ainv + Ainv * B * T * C * Ainv,
                         - Ainv * B * T,
                         - T * C * Ainv,
                           T );
    }

    M_HOST void print() {
        printf("[ %.4f + %.4fj\t", A.a.real(), A.a.imag()); printf("%.4f + %.4fj\t", A.b.real(), A.b.imag()); printf("%.4f + %.4fj\t", B.a.real(), B.a.imag()); printf("%.4f + %.4fj  \n", B.b.real(), B.b.imag());
        printf("  %.4f + %.4fj\t", A.c.real(), A.c.imag()); printf("%.4f + %.4fj\t", A.d.real(), A.d.imag()); printf("%.4f + %.4fj\t", B.c.real(), B.c.imag()); printf("%.4f + %.4fj  \n", B.d.real(), B.d.imag());
        printf("  %.4f + %.4fj\t", C.a.real(), C.a.imag()); printf("%.4f + %.4fj\t", C.b.real(), C.b.imag()); printf("%.4f + %.4fj\t", D.a.real(), D.a.imag()); printf("%.4f + %.4fj  \n", D.b.real(), D.b.imag());
        printf("  %.4f + %.4fj\t", C.c.real(), C.c.imag()); printf("%.4f + %.4fj\t", C.d.real(), C.d.imag()); printf("%.4f + %.4fj\t", D.c.real(), D.c.imag()); printf("%.4f + %.4fj ]\n", D.d.real(), D.d.imag());
    }

    static M_HOSTDEVICE matrix4_4 zero(void) {
        return matrix4_4(matrix2_2::zero(), matrix2_2::zero(),
                         matrix2_2::zero(), matrix2_2::zero());
    }

    static M_HOSTDEVICE matrix4_4 ones(void) {
        return matrix4_4(matrix2_2::ones(), matrix2_2::ones(),
                         matrix2_2::ones(), matrix2_2::ones());
    }

    static M_HOSTDEVICE matrix4_4 unit(void) {
        return matrix4_4(matrix2_2::unit(), matrix2_2::zero(),
                         matrix2_2::zero(), matrix2_2::unit());
    }
};

template <int height> struct matrixstack4_4
{
    matrixstack2_2<height> A, B, C, D;

    M_HOSTDEVICE matrix4_4 operator[](int index) const {
        return matrix4_4(A[index],B[index],
                         C[index],D[index]);
    }

    M_HOSTDEVICE void set(int index, const matrix4_4 REF(data))
    {
        A.set(index, data.A.a, data.A.b, data.A.c, data.A.d);
        B.set(index, data.B.a, data.B.b, data.B.c, data.B.d);
        C.set(index, data.C.a, data.C.b, data.C.c, data.C.d);
        D.set(index, data.D.a, data.D.b, data.D.c, data.D.d);
    }

    M_HOSTDEVICE void set(int index, const matrix2_2 REF(_A), const matrix2_2 REF(_B), const matrix2_2 REF(_C), const matrix2_2 REF(_D))
    {
        A.set(index, _A.a, _A.b, _A.c, _A.d);
        B.set(index, _B.a, _B.b, _B.c, _B.d);
        C.set(index, _C.a, _C.b, _C.c, _C.d);
        D.set(index, _D.a, _D.b, _D.c, _D.d);
    }

    M_HOSTDEVICE void set(int index,
        const cudacomplex REF(Aa), const cudacomplex REF(Ab), const cudacomplex REF(Ac), const cudacomplex REF(Ad),
        const cudacomplex REF(Ba), const cudacomplex REF(Bb), const cudacomplex REF(Bc), const cudacomplex REF(Bd),
        const cudacomplex REF(Ca), const cudacomplex REF(Cb), const cudacomplex REF(Cc), const cudacomplex REF(Cd),
        const cudacomplex REF(Da), const cudacomplex REF(Db), const cudacomplex REF(Dc), const cudacomplex REF(Dd)
    )
    {
        A.set(index, Aa, Ab, Ac, Ad);
        B.set(index, Ba, Bb, Bc, Bd);
        C.set(index, Ca, Cb, Cc, Cd);
        D.set(index, Da, Db, Dc, Dd);
    }
};

__global__ void
testKernel( matrixstack4_4<STACKHEIGHT>* g_idata, matrixstack4_4<STACKHEIGHT>* g_odata) 
{
  const unsigned int tid    = threadIdx.x;
  const unsigned int sample = blockIdx.x;

  matrix4_4 tmp = g_idata[sample][tid] * g_idata[sample][tid].inv();

  g_odata[sample].set(tid, tmp.A.a, tmp.A.b, tmp.A.c, tmp.A.d,
                           tmp.B.a, tmp.B.b, tmp.B.c, tmp.B.d,
                           tmp.C.a, tmp.C.b, tmp.C.c, tmp.C.d,
                           tmp.D.a, tmp.D.b, tmp.D.c, tmp.D.d);
}

void runTest( int argc, char** argv);

int main( int argc, char** argv) 
{
    runTest( argc, argv);
}

void
runTest( int argc, char** argv) 
{
    cudaSetDevice(0);
    cudaFree(NULL);

    unsigned int num_samples = 1000;
    unsigned int mem_size = sizeof( matrixstack4_4<STACKHEIGHT> ) * num_samples;

    printf("Complex class and Complex 4x4 matrix class functional test: Performing\n");
    printf("%d complex 4x4 matrix inversions and multiplications (in %d bytes)\n", STACKHEIGHT * num_samples, mem_size);

    matrixstack4_4<STACKHEIGHT>* h_idata = (matrixstack4_4<STACKHEIGHT>*) malloc( mem_size );
    for( unsigned int i = 0; i < num_samples; ++i) 
    {
        for ( unsigned int j = 0; j < STACKHEIGHT; j++)
        {
            cudacomplex Aa = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Ab = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Ac = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Ad = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            matrix2_2 A(Aa,Ab,Ac,Ad);
            cudacomplex Ba = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Bb = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Bc = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Bd = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            matrix2_2 B(Ba,Bb,Bc,Bd);
            cudacomplex Ca = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Cb = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Cc = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Cd = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            matrix2_2 C(Ca,Cb,Cc,Cd);
            cudacomplex Da = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Db = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Dc = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            cudacomplex Dd = { (float)rand() / RAND_MAX, (float)rand() / RAND_MAX };
            matrix2_2 D(Da,Db,Dc,Dd);

            h_idata[i].set(j, matrix4_4(A,B,C,D));
        }
    }

    matrixstack4_4<STACKHEIGHT>* d_idata;
    cudaMalloc( (void**) &d_idata, mem_size);
    cudaMemcpy( d_idata, h_idata, mem_size, cudaMemcpyHostToDevice);

    matrixstack4_4<STACKHEIGHT>* d_odata;
    cudaMalloc( (void**) &d_odata, mem_size);
    cudaMemcpy( d_odata, h_idata, mem_size, cudaMemcpyHostToDevice);

    dim3  grid( num_samples, 1, 1);
    dim3  threads( STACKHEIGHT, 1, 1);
    testKernel<<< grid, threads >>>( d_idata, d_odata );
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) {
        fprintf(stderr, "CUDA error: Kernel execution failed in file <%s>, line %i : %s.\n",
                __FILE__, __LINE__, cudaGetErrorString( err) );
        exit(-1);
    }

    matrixstack4_4<STACKHEIGHT>* h_odata = (matrixstack4_4<STACKHEIGHT>*) malloc( mem_size );
    cudaMemcpy( h_odata, d_odata, mem_size,
                                cudaMemcpyDeviceToHost);

    free( h_idata);
    free( h_odata);
    cudaFree(d_idata);
    cudaFree(d_odata);

    cudaThreadExit();
}
